@@ -35,11 +35,16 @@
 #include <sys/socket.h>
 #include <netinet/in.h>
 
+#include <sys/types.h>
+#include <sys/socket.h>
+#include <netinet/tcp.h>
+
 #include "common.h"
 #include "server.h"
 #include "connection.h"
 #include "log.h"
 
+
 static void purge_closed_connection       (cherokee_server_t *srv, cherokee_connection_t *conn);
 static void maybe_purge_closed_connection (cherokee_server_t *srv, cherokee_connection_t *conn);
 
@@ -77,10 +82,15 @@ cherokee_server_new  (cherokee_server_t **srv)
 	/* Server config
 	 */
 	n->port           = 80;
+	n->listen_to      = NULL;
+	n->fdwatch_msecs  = 999;
 	n->log            = 0;
 	n->timeout        = 15;
 	n->keepalive      = 1;
+	n->hideversion    = 0;
 	n->mimetypes_file = NULL;
+	n->userdir        = NULL;
+	n->userdir_handler= NULL;
 
 	n->user_orig      = getuid();
 	n->user           = n->user_orig;
@@ -114,9 +124,9 @@ change_execution_user (cherokee_server_t *srv)
 	 */
 	if (srv->group != srv->group_orig) {
 		error = setgid (srv->group);
-		if (error) {
-			PRINT_ERROR ("Can't change group: setgid(\"%d\") failed\n", srv->group);
-			return ret_error;
+		if (error != 0) {
+			PRINT_ERROR ("Can't change group to GID %d, running with GID=%d\n",
+				     srv->group, srv->group_orig);
 		}
 	}
 
@@ -124,15 +134,23 @@ change_execution_user (cherokee_server_t *srv)
 	 */
 	if (srv->user != srv->user_orig) {
 		error = setuid (srv->user);		
-		if (error) {
-			PRINT_ERROR ("Can't change user: setuid(\"%d\") failed\n", srv->user);
-			return ret_error;
-		}			
+		if (error != 0) {
+			PRINT_ERROR ("Can't change user to UID %d, running with UID=%d\n",
+				     srv->user, srv->user_orig);
+		}
 	}
 
 	return ret_ok;
 }
 
+
+void  
+cherokee_server_set_min_latency (cherokee_server_t *srv, int msecs)
+{
+	srv->fdwatch_msecs = msecs;
+}
+
+
 static ret_t
 initialize_server_socket (cherokee_server_t *srv)
 {
@@ -168,15 +186,23 @@ initialize_server_socket (cherokee_server_t *srv)
 	 */
 	sockaddr.sin_port        = htons (srv->port);
 	sockaddr.sin_family      = AF_INET;
-	sockaddr.sin_addr.s_addr = INADDR_ANY;
-	
+
+	if (srv->listen_to != NULL) {
+#ifdef HAVE_INET_PTON
+		inet_pton(sockaddr.sin_family, srv->listen_to, &sockaddr.sin_addr);
+#else
+		sockaddr.sin_addr.s_addr = inet_addr(srv->listen_to);
+#endif
+	} else {
+		sockaddr.sin_addr.s_addr = INADDR_ANY;
+	}
+
 	ret = bind (srv->socket, (struct sockaddr *)&sockaddr, sizeof(sockaddr));
 	if (ret != 0) {
 		uid_t uid = getuid();
 		gid_t gid = getgid();
 
-		cherokee_log (LOG_ERR, "Can't bind() socket (port %d, uid %d, %gid)\n", srv->port, uid, gid);
-		fprintf (stderr, "Can't bind() socket (port %d, uid %d, gid %d)\n", srv->port, uid, gid);
+		PRINT_ERROR ("Can't bind() socket (port=%d, UID=%d, GID=%d)\n", srv->port, uid, gid);
 		return ret_error;
 	}
 
@@ -267,11 +293,26 @@ cherokee_server_clean (cherokee_server_t *srv)
 
 	/* Clean the configuration
 	 */
+	if (srv->listen_to != NULL) {
+		free (srv->listen_to);
+		srv->listen_to = NULL;
+	}
+
 	if (srv->mimetypes_file != NULL) {
 		free (srv->mimetypes_file);
 		srv->mimetypes_file = NULL;
 	}
 
+	if (srv->userdir != NULL) {
+		free (srv->userdir);
+		srv->userdir = NULL;
+	}
+	if (srv->userdir_handler != NULL) {
+		cherokee_plugin_table_entry_free (srv->userdir_handler);
+		srv->userdir_handler = NULL;
+	}
+
+
 	/* Clean the virtual servers table
 	 */
 	cherokee_virtual_server_clean (srv->vserver_default);
@@ -281,6 +322,8 @@ cherokee_server_clean (cherokee_server_t *srv)
 	cherokee_table_free (srv->vservers);
 	cherokee_table_new (&srv->vservers);
 
+	srv->hideversion = 0;
+
 	return ret_ok;
 }
 
@@ -296,11 +339,30 @@ cherokee_server_free (cherokee_server_t *srv)
 	cherokee_virtual_server_free (srv->vserver_default);
 	srv->vserver_default = NULL;
 
+	// TODO:
+	// Free each vserver
+	cherokee_table_free (srv->vservers);
+	cherokee_table_new (&srv->vservers);
+
+
+	if (srv->listen_to != NULL) {
+		free (srv->listen_to);
+		srv->listen_to = NULL;
+	}
+
 	if (srv->mimetypes_file != NULL) {
 		free (srv->mimetypes_file);
 		srv->mimetypes_file = NULL;
 	}
 
+	if (srv->userdir != NULL) {
+		free (srv->userdir);
+		srv->userdir = NULL;
+	}
+	if (srv->userdir_handler != NULL) {
+		cherokee_plugin_table_entry_free (srv->userdir_handler);
+		srv->userdir_handler = NULL;
+	}
 	
 	free (srv);
 	
@@ -340,20 +402,16 @@ look_for_new_connection (cherokee_server_t *srv)
 	 */
 	if (cherokee_fdpoll_check (srv->fdpoll, srv->socket, 0))
 	{
+		int tmp = 1;
 		int new_socket;
 		cherokee_connection_t *new_connection;
 		
-		
-		/* Get the new socket connection
-		 */
-		new_socket = accept (srv->socket, NULL, NULL);
-		cherokee_fdpoll_add (srv->fdpoll, new_socket, 0);
 
-		
 		/* Look for a reusable connection
 		 */
 		if (list_empty (&srv->reuse_list)) {
 			cherokee_connection_new (&new_connection);
+			new_connection->server = srv;
 		} else {
 			new_connection = CONN(srv->reuse_list.next);
 			list_del ((list_t *)new_connection);
@@ -362,6 +420,37 @@ look_for_new_connection (cherokee_server_t *srv)
 		}
 
 		
+		/* Get the new socket connection
+		 */
+		if (srv->log) {
+			socklen_t len; 
+			len = sizeof(struct sockaddr_in);
+			new_socket = accept (srv->socket, &new_connection->addr_in, &len);
+		} else {
+			new_socket = accept (srv->socket, NULL, NULL);
+		}
+		cherokee_fdpoll_add (srv->fdpoll, new_socket, 0);
+
+
+		/*
+		 *  EXPERIMENTAL
+		 *
+		 */
+
+		fcntl (new_socket, F_SETFD, 1);  /* close-on-exec */
+		
+		/* Disable Nagle's algorithm for this connection. 
+		 * Written data to the network is not buffered pending 
+		 * acknowledgement of previously written data.
+		 */
+		setsockopt (new_socket, IPPROTO_TCP, TCP_NODELAY, &tmp, sizeof(tmp));
+
+		/*
+		 *  EXPERIMENTAL
+		 *
+		 */
+
+		
 		/* Set the socket
 		 */
 		cherokee_connection_set_socket (new_connection, new_socket);
@@ -371,7 +460,6 @@ look_for_new_connection (cherokee_server_t *srv)
 		 */
 		list_add ((list_t *)new_connection, &srv->active_list);
 
-
 		return 1;
 	}
 	
@@ -382,20 +470,15 @@ look_for_new_connection (cherokee_server_t *srv)
 static inline void
 conn_set_mode (cherokee_server_t *srv, cherokee_connection_t *conn, cherokee_connection_status_t rw)
 {
-	ret_t ret;
-
 	conn->status = rw;
-	ret = cherokee_fdpoll_set_mode (srv->fdpoll, conn->socket, rw);
-	if (ret != ret_ok) {
-		SHOULDNT_HAPPEN;
-	}
+	cherokee_fdpoll_set_mode (srv->fdpoll, conn->socket, rw);
 }
 
 
 static void
 purge_closed_connection (cherokee_server_t *srv, cherokee_connection_t *conn)
 {
-	/* Delete from File descriptors poll
+	/* Delete from file descriptors poll
 	 */
 	cherokee_fdpoll_del (srv->fdpoll, conn->socket);
 	
@@ -420,10 +503,10 @@ maybe_purge_closed_connection (cherokee_server_t *srv, cherokee_connection_t *co
 {
 	/* Maybe it's a keep-alive connection
 	 */
-	if (conn->keep_alive) {
+	if (conn->keep_alive != 0) {
 		cherokee_connection_clean (conn);
-		conn->timeout = srv->bogo_now + srv->timeout;
 		conn_set_mode (srv, conn, stat_reading);
+		conn->timeout = srv->bogo_now + srv->timeout;
 		return;
 	}
 
@@ -443,7 +526,7 @@ process_active_connections (cherokee_server_t *srv)
 	 */
 	list_for_each_safe (i, tmp, (list_t*)&srv->active_list) {
 		conn = CONN(i);			 											   
-		
+
 		/* May the connection was too much time w/o any work
 		 */
 		if (conn->timeout < srv->bogo_now) {
@@ -461,12 +544,11 @@ process_active_connections (cherokee_server_t *srv)
 		 */
 		conn->timeout = srv->bogo_now + srv->timeout;
 
-		
-		/* Reading
+		/* Phases
 		 */
-		switch (conn->status) {
-		case stat_reading:
-		{	
+		switch (conn->phase) {
+		case phase_reading_header:
+
 			ret = cherokee_connection_recv (conn);
 			if (ret != ret_ok) {
 				purge_closed_connection (srv, conn);
@@ -478,170 +560,187 @@ process_active_connections (cherokee_server_t *srv)
 			ret = cherokee_connection_reading_check (conn);
 			if (ret != ret_ok) {
 				conn_set_mode (srv, conn, stat_writing);
+				conn->phase = phase_add_headers;
 				continue;
 			}
 
 			/* May it already has the full header
 			 */
 			if (cherokee_connection_eoh (conn)) {
-				cherokee_virtual_server_t *vserver = NULL;
-
+				conn->phase = phase_processing_header;
+				
 				/* Turn the connection in write mode
 				 */
 				conn_set_mode (srv, conn, stat_writing);
-				
-				/* Get the request
-				 */
-				ret = cherokee_connection_get_request (conn);
-				if (ret != ret_ok) {
-					continue;
-				}
+			} else 
+				break;
 
-				/* Maybe log it
-				 */
-				cherokee_log (LOG_INFO, conn->request);
-
-				/* Server's "Keep-Alive" can be turned "Off"
-				 */
-				conn->keep_alive &= srv->keepalive;
-
-				/* Get the virtual host
-				 */
-				if (conn->host) {
-					vserver = cherokee_table_get (srv->vservers, conn->host);
-				}
-				if (vserver == NULL) {
-					vserver = srv->vserver_default;
-				}
-
-				/* Build local request string
-				 */
-				ret = cherokee_connection_build_local_request (conn, vserver);
-				if (ret != ret_ok) {
-					continue;
-				}
-
-				/* Create the handler
-				 */
-				ret = cherokee_connection_create_handler (conn, vserver->plugins);
-				
-				if (ret != ret_ok) {
-					continue;
-				}				
-
-				/* Parse the rest of headers
-				 */
-				ret = cherokee_connection_parse_header (conn);
-				if (ret != ret_ok) {
-					continue;
-				}
-
-				/* Look for the request
-				 */
-				ret = cherokee_connection_open_request (conn);
-				if (ret != ret_ok) {
-					continue;
-				}
-			}			
-		}
-		break;
-		
-		/* Writing
-		 */
-		case stat_writing:
-		{
-			/* Header
+		case phase_processing_header: {
+			
+			cherokee_virtual_server_t *vserver = NULL;
+			cherokee_plugin_table_entry_t *plugin_entry = NULL;
+			
+			/* Get the request
 			 */
-			if (conn->header_sent == 0) {
-				ret = cherokee_connection_send_header (conn);
-				
-				if (ret == ret_ok) {
-					conn->header_sent = 1;
-					
-					if (conn->method == http_head) {
-						maybe_purge_closed_connection (srv, conn);
-						continue;
-					}
-					
-					if (! http_type_200(conn->error_code)) {
-						cherokee_connection_send_response_page (conn);
-						purge_closed_connection (srv, conn);
-						continue;
-					}
-				}
-				
-			/* Body
+			ret = cherokee_connection_get_request (conn);
+			if (ret != ret_ok) {
+				conn->phase = phase_add_headers;
+				continue;
+			}
+			
+			/* Maybe log it
 			 */
+			if (srv->log) {
+				cherokee_connection_log (conn, srv->bogo_now);
+			}
+
+			
+			/* Get the virtual host
+			 */
+			if (conn->host) {
+				vserver = cherokee_table_get (srv->vservers, conn->host);
+			}
+			if (vserver == NULL) {
+				vserver = srv->vserver_default;
+			}
+			
+			/* Build local request string
+			*/
+			if ((srv->userdir != NULL) && 
+			    (srv->userdir_handler != NULL) && 
+			    (cherokee_connection_is_userdir (conn)))
+			{
+				ret = cherokee_connection_build_local_directory_userdir (conn, srv->userdir);
+				plugin_entry = srv->userdir_handler;
+				
 			} else {
-				ret_t sent_ret;
+				ret = cherokee_connection_build_local_directory (conn, vserver);
+				if (ret != ret_ok) {
+					conn->phase = phase_add_headers;
+					continue;
+				}
+				ret = cherokee_connection_get_plugin_entry (conn, vserver->plugins, &plugin_entry);
+			}
+			if (ret != ret_ok) {
+				conn->phase = phase_add_headers;
+				continue;
+			}
+			
+			ret = cherokee_connection_create_handler (conn, plugin_entry);
+			if (ret != ret_ok) {
+				conn->phase = phase_add_headers;
+				continue;
+			}
+			
+			
+			/* Parse the rest of headers
+			 */
+			ret = cherokee_connection_parse_header (conn);
+			if (ret != ret_ok) {
+				conn->phase = phase_add_headers;
+				continue;
+			}
 
-				/* Handler step: read or make new data to send
-				 */
-				ret = cherokee_connection_step (conn);
+			conn->phase = phase_init;
+		}
 
-				switch (ret) {
-				case ret_ok:
-				case ret_eof_have_data:
-					/* We've data, so..
-					 */
-					sent_ret = cherokee_connection_send (conn);
-					if (sent_ret == ret_eof) {
-						purge_closed_connection (srv, conn);
-					} else if (ret == ret_eof_have_data) {
-						maybe_purge_closed_connection (srv, conn);
-					}
-					break;
-				case ret_eof:
+		case phase_init: {
+
+			/* Server's "Keep-Alive" can be turned "Off"
+			 */
+			if (srv->keepalive == 0) {
+				conn->keep_alive = 0;
+			}
+
+			/* Look for the request
+			 */
+			ret = cherokee_connection_open_request (conn);
+			if (ret != ret_eagain) {
+				conn->phase = phase_add_headers;
+			}
+
+		} break;
+			
+		case phase_add_headers:
+			ret = cherokee_connection_send_header (conn, srv->hideversion);
+				
+			if (ret == ret_ok) {				
+				if (conn->method == http_head) {
 					maybe_purge_closed_connection (srv, conn);
-					break;
-				case ret_error:
+					continue;
+				}
+				
+				if (! http_type_200(conn->error_code)) {
+					cherokee_connection_send_response_page (conn, srv->hideversion);
 					purge_closed_connection (srv, conn);
-					break;
-				default:
-					SHOULDNT_HAPPEN;
+					continue;
 				}
 			}
-		}
-		break;
-		
-		default:
-			SHOULDNT_HAPPEN;
+
+			conn->phase = phase_steping;
 			
-		} /* switch (conn->staus) */
-	} /* list_for_each */
+		case phase_steping: {
+			ret_t sent_ret;
+			
+			/* Handler step: read or make new data to send
+			 */
+			ret = cherokee_connection_step (conn);
+			
+			switch (ret) {
+			case ret_eof_have_data:
+				sent_ret = cherokee_connection_send (conn);
+
+				switch (sent_ret) {
+				case ret_error:
+					purge_closed_connection (srv, conn);
+					continue;
+				default:	
+					maybe_purge_closed_connection (srv, conn);
+					continue;
+				}
+				break;
+
+			case ret_ok:
+				/* We've data, so..
+				 */
+				sent_ret = cherokee_connection_send (conn);
+
+				if ((sent_ret == ret_eof) ||
+				    (sent_ret == ret_error)) 
+				{
+					purge_closed_connection (srv, conn);					
+				}
+				break;
+
+			case ret_eof:
+				maybe_purge_closed_connection (srv, conn);
+				break;
+			case ret_error:
+				purge_closed_connection (srv, conn);
+				break;
+			case ret_eagain:
+				printf ("eagain en servidor\n");
+				break;
+			default:
+				SHOULDNT_HAPPEN; 
+			}
+			
+		} break;
+
+		default:
+ 			SHOULDNT_HAPPEN;
+		}
+	} /* list */
 }
 
 
-/*
-ret_t 
-cherokee_server_set (cherokee_server_t *srv, char *dir, char *handler)
-{
-	   ret_t ret;
-
-	   ret = cherokee_plugin_table_set (srv->plugins, dir, handler);
-
-	   if (ret != ret_ok) {
-			 fprintf (stderr, "Can't load handler '%s': ", handler);
-			 switch (ret) {
-			 case ret_file_not_found:
-				    fprintf (stderr, "file not found\n");
-				    break;
-			 default:
-				    fprintf (stderr, "unknown error\n"); 
-			 }
-	   }
-
-	   return ret;
-}
-*/
-
 void
 cherokee_server_step (cherokee_server_t *srv)
 {
 	int ret, i=MAX_NEW_CONNECTIONS_PER_STEP;
 	
 	do {
-		cherokee_fdpoll_watch (srv->fdpoll, 999);
+		cherokee_fdpoll_watch (srv->fdpoll, srv->fdwatch_msecs);
 
 		/* Look for new connections
 		 */	   
@@ -687,5 +786,18 @@ cherokee_server_read_config_file (cherokee_server_t *srv, char *filename)
 ret_t 
 cherokee_server_read_config_string (cherokee_server_t *srv, char *config_string)
 {
+	int   error;
+	void *bufstate;
+
+	bufstate = (void *) yy_scan_string (config_string);
+	yy_switch_to_buffer(bufstate);
+
+	error = yyparse((void *)srv);
+
+	yy_delete_buffer (bufstate);
+
+	if (error != 0)
+		return ret_error;
+
 	return ret_ok;
 }
