@@ -23,8 +23,8 @@
  * USA
  */
 
-#ifndef __CHEROKEE2_SERVER_H__
-#define __CHEROKEE2_SERVER_H__
+#ifndef __CHEROKEE_SERVER_H__
+#define __CHEROKEE_SERVER_H__
 
 #include <time.h>
 
@@ -53,6 +53,7 @@ typedef struct {
 	/* Configuration
 	 */
 	int   socket;
+	char *listen_to;
 
 	int   fdwatch_msecs;
 
@@ -60,6 +61,7 @@ typedef struct {
 	int   log;
 	int   timeout;
 	int   keepalive;
+	int   hideversion;
 
 	uid_t user;
 	uid_t user_orig;
@@ -89,4 +91,4 @@ ret_t cherokee_server_log   (cherokee_server_t  *srv, int active);
 ret_t cherokee_server_read_config_file   (cherokee_server_t  *srv, char *filename);
 ret_t cherokee_server_read_config_string (cherokee_server_t  *srv, char *string);
 
-#endif /* __CHEROKEE2_SERVER_H__ */
+#endif /* __CHEROKEE_SERVER_H__ */
